// -*- C++ -*-
//
// Package:    ETH_digiDumper
// Class:      ETH_digiDumper
// 
/**\class ETH_digiDumper ETH_digiDumper.cc Tools/ETH_digiDumper/src/ETH_digiDumper.cc

 Description: Simple class to read and store the mgpa digis from ECAL local runs

 Implementation:
     <Notes on implementation>
*/
//
// Original Author:  Konstantinos Theofilatos
//         Created:  Wed Mar 10 19:22:09 CET 2010
// $Id$
//
//


// system include files
#include <memory>

// user include files
#include "FWCore/Framework/interface/Frameworkfwd.h"
#include "FWCore/Framework/interface/EDAnalyzer.h"

#include "FWCore/Framework/interface/Event.h"
#include "FWCore/Framework/interface/MakerMacros.h"

#include "FWCore/ParameterSet/interface/ParameterSet.h"

#include "FWCore/ServiceRegistry/interface/Service.h"
#include "PhysicsTools/UtilAlgos/interface/TFileService.h"

#include "DataFormats/EcalDigi/interface/EcalDigiCollections.h"
#include "DataFormats/EcalRawData/interface/EcalRawDataCollections.h"

#include "DataFormats/DetId/interface/DetId.h"
#include "DataFormats/EcalDetId/interface/EBDetId.h"
#include "DataFormats/EcalDetId/interface/EEDetId.h"

#include "TTree.h"
//
// class decleration
//

class ETH_digiDumper : public edm::EDAnalyzer {
   public:
      explicit ETH_digiDumper(const edm::ParameterSet&);
      ~ETH_digiDumper();


   private:
      virtual void beginJob() ;
      virtual void analyze(const edm::Event&, const edm::EventSetup&);
      virtual void endJob() ;

      edm::Service<TFileService> fTFileService_;

      edm::InputTag EBDigiCollection_;
      edm::InputTag EEDigiCollection_;
      edm::Handle<EBDigiCollection> EBdigis_;
      edm::Handle<EEDigiCollection> EEdigis_;

      TTree *tree_;
      float sampleTime_[10];
      float mgpaSample_[10];
      int sampleGainId_[10];
      int iphi_;
      int ieta_;
      int iz_;
      int subDet_;


      // ----------member data ---------------------------
};

//
// constants, enums and typedefs
//

//
// static data member definitions
//

//
// constructors and destructor
//
ETH_digiDumper::ETH_digiDumper(const edm::ParameterSet& iConfig)

{
   //now do what ever initialization is needed
    EBDigiCollection_ = iConfig.getParameter<edm::InputTag>("EBDigiCollection");
    EEDigiCollection_ = iConfig.getParameter<edm::InputTag>("EEDigiCollection");

    tree_ = fTFileService_->make<TTree>("analysis","analysis");
    tree_->Branch("sampleTime",sampleTime_,"sampleTime[10]/F");
    tree_->Branch("mgpaSample",mgpaSample_,"mgpaSample[10]/F");
    tree_->Branch("sampleGainId",sampleGainId_,"sampleGainId[10]/I");

    tree_->Branch("iphi",&iphi_,"iphi/I");
    tree_->Branch("ieta",&ieta_,"ieta/I");
    tree_->Branch("iz",&iz_,"iz/I");
    tree_->Branch("subDet",&subDet_,"subDet/I");


}


ETH_digiDumper::~ETH_digiDumper()
{
 
   // do anything here that needs to be done at desctruction time
   // (e.g. close files, deallocate resources etc.)

}


//
// member functions
//

// ------------ method called to for each event  ------------
void
ETH_digiDumper::analyze(const edm::Event& iEvent, const edm::EventSetup& iSetup)
{
   using namespace edm;
  iEvent.getByLabel(EBDigiCollection_, EBdigis_);
  iEvent.getByLabel(EEDigiCollection_, EEdigis_);

  for(EBDigiCollection::const_iterator It = EBdigis_->begin(); It != EBdigis_->end(); ++It)
  { 

    EBDetId ebDetId  = (*It).id();
    EcalDataFrame df;
    df = (*It);

    int iphi = ebDetId.iphi();
    int ieta = ebDetId.ieta();
    int iz = ebDetId.zside();
    int subDet = 0;

    subDet_ = subDet;

    iphi_ = iphi;
    ieta_ = ieta;
    iz_ = iz;

    for(int iSample = 0; iSample < 10; iSample++)
    { 

      mgpaSample_[iSample] = float(df.sample(iSample).adc());
      sampleGainId_[iSample] = df.sample(iSample).gainId();
      sampleTime_[iSample] = iSample;
    }
    
    tree_->Fill();
  }


#ifdef THIS_IS_AN_EVENT_EXAMPLE
   Handle<ExampleData> pIn;
   iEvent.getByLabel("example",pIn);
#endif
   
#ifdef THIS_IS_AN_EVENTSETUP_EXAMPLE
   ESHandle<SetupData> pSetup;
   iSetup.get<SetupRecord>().get(pSetup);
#endif
}


// ------------ method called once each job just before starting event loop  ------------
void 
ETH_digiDumper::beginJob()
{
}

// ------------ method called once each job just after ending the event loop  ------------
void 
ETH_digiDumper::endJob() {
}

//define this as a plug-in
DEFINE_FWK_MODULE(ETH_digiDumper);
